# -*- coding: utf-8 -*-
"""
.. Authors
    Novimir A. Pablant <npblant@pppl.gov>

Contains a set of "external" libraries and tools that are included in the
XICSRT source code. These are included here either because they are not
currently available on pipy or because they have been highly modified.

Programming Notes
-----------------

Make sure not to add anything here that is incompatible with the MIT license!
"""
