"""
Version Description:
major.minor.revision

major:
  - Major changes in codebase or functionality.

minor:
  - Changes to the public API
  - Old config files can no longer be loaded without conversion.
  - New output dictionaries may break existing external code.

revision:
  - New release that does not break any compatibility.
"""

__version__="0.8.12"
__version_date__ = "2024-08-20"
